from __future__ import annotations

import logging
from typing import Dict, Optional

import pandas as pd
import spectrum_fundamentals.constants as c
from spectrum_fundamentals.mod_string import internal_without_mods

from .search_results import SearchResults, parse_mods

logger = logging.getLogger(__name__)


class Sage(SearchResults):
    """Handle search results from Sage."""

    @property
    def standard_mods(self):
        """Standard modifications that are always applied if not otherwise specified."""
        return {"C[+57.0215]": 4, "M[+15.9949]": 35, "M[+15.994]": 35}

    def read_result(
        self,
        tmt_label: str = "",
        custom_mods: dict[str, int] | None = None,
        ptm_unimod_id: int | None = 0,
        ptm_sites: list[str] | None = None,
    ) -> pd.DataFrame:
        """
        Function to read a msms tsv and perform some basic formatting.

        :param tmt_label: optional tmt label as str
        :param custom_mods: optional dictionary mapping Sage-specific mod pattern to UNIMOD IDs.
            If None, static carbamidomethylation of cytein and variable oxidation of methionine
            are mapped automatically. To avoid this, explicitely provide an empty dictionary.
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        :return: pd.DataFrame with the formatted data
        """
        parsed_mods = parse_mods(self.standard_mods | (custom_mods or {}))
        if tmt_label:
            unimod_tag = c.TMT_MODS[tmt_label]
            parsed_mods[r"K\[\+\d+\.\d+\]"] = f"K{unimod_tag}"
            parsed_mods[r"^\[\+\d+\.\d+\]"] = f"{unimod_tag}"

        logger.info(f"Reading {self.path}")
        self.results = pd.read_csv(
            self.path,
            usecols=["filename", "scannr", "peptide", "charge", "hyperscore", "calcmass", "label", "proteins"],
            sep="\t",
        )
        logger.info(f"Finished reading {self.path}")

        self.convert_to_internal(mods=parsed_mods, ptm_unimod_id=ptm_unimod_id, ptm_sites=ptm_sites)
        return self.filter_valid_prosit_sequences()

    def filter_valid_prosit_sequences(self):
        """Filter valid Prosit sequences."""
        logger.info(f"#sequences before filtering for valid prosit sequences: {len(self.results.index)}")
        # retain only peptides that fall within [7, 30] length supported by Prosit
        self.results = self.results[(self.results["PEPTIDE_LENGTH"] <= 30) & (self.results["PEPTIDE_LENGTH"] >= 7)]
        # remove unsupported mods to exclude
        self.results = self.results[~self.results["MODIFIED_SEQUENCE"].str.contains(r"\[\d+\]", regex=True)]
        # remove precursor charges greater than 6
        self.results = self.results[self.results["PRECURSOR_CHARGE"] <= 6]
        logger.info(f"#sequences after filtering for valid prosit sequences: {len(self.results.index)}")

        return self.results

    def convert_to_internal(self, mods: dict[str, str], ptm_unimod_id: int | None, ptm_sites: list[str] | None):
        """
        Convert all columns in the Sage output to the internal format used by Oktoberfest.

        :param mods: dictionary mapping Sage-specific mod patterns (keys) to ProForma standard (values)
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        """
        df = self.results

        df.fillna({"proteins": "UNKNOWN"}, inplace=True)
        df.replace({"filename": {r"\.mz[M|m][l|L]": ""}, "peptide": mods}, regex=True, inplace=True)
        if not df["scannr"].dtype == int:
            df["scannr"] = df["scannr"].str.rsplit(pat="=", n=1).str[1].astype("int64")
        df["label"] = df["label"] < 0
        df["SEQUENCE"] = internal_without_mods(df["peptide"])
        df["PEPTIDE_LENGTH"] = df["SEQUENCE"].str.len()

        df.rename(
            columns={
                "filename": "RAW_FILE",
                "scannr": "SCAN_NUMBER",
                "peptide": "MODIFIED_SEQUENCE",
                "charge": "PRECURSOR_CHARGE",
                "calcmass": "MASS",
                "hyperscore": "SCORE",
                "label": "REVERSE",
                "proteins": "PROTEINS",
            },
            inplace=True,
        )



from __future__ import annotations

import logging
from pathlib import Path
from typing import Dict, Optional, Union

import pandas as pd
import spectrum_fundamentals.constants as c
from spectrum_fundamentals.mod_string import add_permutations, internal_without_mods

from .search_results import SearchResults, parse_mods

logger = logging.getLogger(__name__)


class MaxQuant(SearchResults):
    """Handle search results from MaxQuant."""

    def __init__(self, path: str | Path):
        """
        Init Searchresults object.

        :param path: path to file
        """
        if isinstance(path, str):
            path = Path(path)

        if path.is_file() and path.name == "msms.txt":
            path = path.parent
        self.path = path

    @property
    def standard_mods(self):
        """Standard modifications that are always applied if not otherwise specified."""
        return {
            "C": 4,
            "M(ox)": 35,
            "M(Oxidation (M))": 35,
            "R(Citrullination)": 7,
            "Q(Deamidation (NQ))": 7,
            "N(Deamidation (NQ))": 7,
        }

    @staticmethod
    def add_tmt_mod(mass: float, seq: str, unimod_tag: str) -> float:
        """
        Add tmt modification.

        :param mass: mass without tmt modification
        :param seq: sequence of the peptide
        :param unimod_tag: UNIMOD tag for the modification
        :return: mass as float
        """
        num_of_tmt = seq.count(unimod_tag)
        mass += num_of_tmt * c.MOD_MASSES[f"{unimod_tag}"]
        return mass

    def filter_valid_prosit_sequences(self):
        """Filter valid Prosit sequences."""
        logger.info(f"#sequences before filtering for valid prosit sequences: {len(self.results.index)}")
        # retain only peptides that fall within [7, 30] length supported by Prosit
        self.results = self.results[(self.results["PEPTIDE_LENGTH"] <= 30) & (self.results["PEPTIDE_LENGTH"] >= 7)]
        # remove unsupported mods to exclude
        self.results = self.results[~self.results["MODIFIED_SEQUENCE"].str.contains(r"\(", regex=True)]
        # remove precursor charges greater than 6
        self.results = self.results[self.results["PRECURSOR_CHARGE"] <= 6]
        logger.info(f"#sequences after filtering for valid prosit sequences: {len(self.results.index)}")

        return self.results

    def read_result(
        self,
        tmt_label: str = "",
        custom_mods: dict[str, int] | None = None,
        ptm_unimod_id: int | None = 0,
        ptm_sites: list[str] | None = None,
    ) -> pd.DataFrame:
        """
        Function to read a msms txt and perform some basic formatting.

        :param tmt_label: optional tmt label as str
        :param custom_mods: optional dictionary mapping MaxQuant-specific mod pattern to UNIMOD IDs.
            If None, static carbamidomethylation of cytein and variable oxidation of methionine
            are mapped automatically. To avoid this, explicitely provide an empty dictionary.
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        :return: pd.DataFrame with the formatted data
        """
        parsed_mods = parse_mods(self.standard_mods | (custom_mods or {}))
        if tmt_label:
            unimod_tag = c.TMT_MODS[tmt_label]
            parsed_mods["K"] = f"K{unimod_tag}"
            parsed_mods["^_"] = f"_{unimod_tag}-"

        logger.info("Reading msms.txt file")
        self.results = pd.read_csv(
            self.path / "msms.txt",
            usecols=[
                "Raw file",
                "Scan number",
                "Modified sequence",
                "Charge",
                "Scan event number",
                "Mass",  # = Calculated Precursor mass; TODO get column with experimental Precursor mass instead
                "Score",
                "Reverse",
                "Proteins",
            ],
            sep="\t",
        )

        logger.info("Finished reading msms.txt file")

        self.convert_to_internal(mods=parsed_mods, ptm_unimod_id=ptm_unimod_id, ptm_sites=ptm_sites)
        return self.filter_valid_prosit_sequences()

    def convert_to_internal(self, mods: dict[str, str], ptm_unimod_id: int | None, ptm_sites: list[str] | None):
        """
        Convert all columns in the MaxQuant output to the internal format used by Oktoberfest.

        :param mods: dictionary mapping MaxQuant-specific mod patterns (keys) to ProForma standard (values)
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        """
        df = self.results
        # Standardize column names
        # df.columns = df.columns.str.upper()
        # df.columns = df.columns.str.replace(" ", "_")
        # df.rename(columns={"CHARGE": "PRECURSOR_CHARGE"}, inplace=True)

        mods["_"] = ""

        df.fillna({"Reverse": "", "Proteins": "UNKNOWN"}, inplace=True)
        df["Reverse"] = df["Reverse"].astype(bool)
        df.replace({"Modified sequence": mods}, regex=True, inplace=True)

        df["Sequence"] = internal_without_mods(df["Modified sequence"])
        df["PEPTIDE_LENGTH"] = df["Sequence"].str.len()
        if ptm_unimod_id != 0:

            # PTM permutation generation
            if ptm_unimod_id == 7:
                allow_one_less_modification = True
            else:
                allow_one_less_modification = False

            df["Modified sequence"] = df["Modified sequence"].apply(
                add_permutations,
                unimod_id=ptm_unimod_id,
                residues=ptm_sites,
                allow_one_less_modification=allow_one_less_modification,
            )
            df = df.explode("Modified sequence", ignore_index=True)

        df.rename(
            columns={
                "Reverse": "REVERSE",
                "Sequence": "SEQUENCE",
                "Modified sequence": "MODIFIED_SEQUENCE",
                "Proteins": "PROTEINS",
                "Charge": "PRECURSOR_CHARGE",
                "Raw file": "RAW_FILE",
                "Scan number": "SCAN_NUMBER",
                "Scan event number": "SCAN_EVENT_NUMBER",
                "Mass": "MASS",
                "Score": "SCORE",
            },
            inplace=True,
        )
        self.results = df

    def generate_internal_timstof_metadata(self):
        """
        Load information files required for correct aggregation of spectra in timsTOF experiments.

        :return: dataframe containing the columns RAW_FILE, SCANNUMBER, PRECURSOR, FRAME, SCANNUMBEGIN, SCANNUMEND, CollisionEnergy
        """
        df_msms = pd.read_csv(self.path / "msms.txt", sep="\t", usecols=["Raw file", "Scan number"])
        df_msms.columns = ["RAW_FILE", "SCAN_NUMBER"]

        df_precursors = pd.read_csv(
            self.path / "accumulatedMsmsScans.txt", sep="\t", usecols=["Raw file", "Scan number", "PASEF precursor IDs"]
        )
        df_precursors.columns = ["RAW_FILE", "SCAN_NUMBER", "PRECURSOR"]
        df_precursors.query("SCAN_NUMBER in @df_msms.SCAN_NUMBER", inplace=True)
        df_precursors["PRECURSOR"] = df_precursors["PRECURSOR"].str.split(";")
        df_precursors = df_precursors.explode("PRECURSOR")
        df_precursors["PRECURSOR"] = df_precursors["PRECURSOR"].astype("int")

        df_pasef = pd.read_csv(
            self.path / "pasefMsmsScans.txt",
            sep="\t",
            usecols=["Raw file", "Frame", "Precursor", "ScanNumBegin", "ScanNumEnd", "CollisionEnergy"],
        )
        df_pasef.columns = ["RAW_FILE", "FRAME", "PRECURSOR", "SCAN_NUM_BEGIN", "SCAN_NUM_END", "COLLISION_ENERGY"]

        return df_pasef.merge(df_precursors).sort_values(["FRAME", "PRECURSOR"])



from __future__ import annotations

import logging
from typing import Dict, Optional

import pandas as pd
from spectrum_fundamentals.constants import PARTICLE_MASSES

from .search_results import SearchResults, parse_mods

logger = logging.getLogger(__name__)


class MSAmanda(SearchResults):
    """Handle search results from MSAmanda."""

    @property
    def standard_mods(self):
        """Standard modifications that are always applied if not otherwise specified."""
        return {"m": 35, "c": 4}

    def read_result(
        self,
        tmt_label: str = "",
        custom_mods: dict[str, int] | None = None,
        ptm_unimod_id: int | None = 0,
        ptm_sites: list[str] | None = None,
        suffix: str = "output.csv",
    ) -> pd.DataFrame:
        """
        Function to read a msms txt and perform some basic formatting.

        :param tmt_label: optional tmt label as str
        :param custom_mods: optional dictionary mapping Sage-specific mod pattern to UNIMOD IDs.
            If None, static carbamidomethylation of cytein and variable oxidation of methionine
            are mapped automatically. To avoid this, explicitely provide an empty dictionary.
        :param suffix: Optional suffix to determine which fileresult files should be taken from the supplied path
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        :raises FileNotFoundError: If the supplied path is not found
        :raises AssertionError: If the supplied path does not contain any files matching the provided suffix.
        :raises NotImplementedError: If tmt label was supplied.
        :return: pd.DataFrame with the formatted data
        """
        parsed_mods = parse_mods(self.standard_mods | (custom_mods or {}))
        print(parsed_mods)
        if tmt_label:
            raise NotImplementedError("TMT is not supported for MSAmanda. Please open an issue on github.")

        if self.path.is_file():
            pathlist = [self.path]
            print("hooray")
        elif self.path.is_dir():
            pathlist = list(self.path.glob(f"*{suffix}"))
            if not pathlist:
                raise AssertionError(f"The directory does not contain any files that match the pattern *{suffix}")
        else:
            raise FileNotFoundError(f"{self.path} does not exist.")

        df_list = []
        for output_file in pathlist:
            logger.info(f"Reading {output_file}...")
            df_list.append(
                pd.read_csv(
                    output_file,
                    sep="\t",
                    skiprows=1,
                    usecols=[
                        "Scan Number",
                        "Sequence",
                        # "Modifications",
                        "Protein Accessions",
                        "Amanda Score",
                        "m/z",
                        "Charge",
                        "Filename",
                    ],
                )
            )
            logger.info(f"Finished reading {output_file}")

        self.results = pd.concat(df_list)

        self.convert_to_internal(mods=parsed_mods, ptm_unimod_id=ptm_unimod_id, ptm_sites=ptm_sites)
        return self.filter_valid_prosit_sequences()

    def filter_valid_prosit_sequences(self):
        """Filter valid Prosit sequences."""
        logger.info(f"#sequences before filtering for valid prosit sequences: {len(self.results.index)}")
        # retain only peptides that fall within [7, 30] length supported by Prosit
        self.results = self.results[(self.results["PEPTIDE_LENGTH"] <= 30) & (self.results["PEPTIDE_LENGTH"] >= 7)]
        # remove unsupported mods to exclude
        self.results = self.results[~self.results["MODIFIED_SEQUENCE"].str.contains(r"[a-z]+", regex=True)]
        # remove precursor charges greater than 6
        self.results = self.results[self.results["PRECURSOR_CHARGE"] <= 6]
        logger.info(f"#sequences after filtering for valid prosit sequences: {len(self.results.index)}")

        return self.results

    def convert_to_internal(self, mods: dict[str, str], ptm_unimod_id: int | None, ptm_sites: list[str] | None):
        """
        Convert all columns in the Sage output to the internal format used by Oktoberfest.

        :param mods: dictionary mapping Sage-specific mod patterns (keys) to ProForma standard (values)
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        """
        df = self.results
        df["REVERSE"] = df["Protein Accessions"].str.startswith("REV_")
        df["m/z"] = df["Charge"] * (df["m/z"] - PARTICLE_MASSES["PROTON"])
        df["SEQUENCE"] = df["Sequence"].str.upper()
        df.replace({"RAW_FILE": {r"\.\w+$": ""}, "Sequence": mods}, regex=True, inplace=True)
        df["PEPTIDE_LENGTH"] = df["SEQUENCE"].str.len()

        df.rename(
            columns={
                "Filename": "RAW_FILE",
                "Scan Number": "SCAN_NUMBER",
                "Sequence": "MODIFIED_SEQUENCE",
                "Charge": "PRECURSOR_CHARGE",
                "m/z": "MASS",
                "Amanda Score": "SCORE",
                "Protein Accessions": "PROTEINS",
            },
            inplace=True,
        )



from __future__ import annotations

import logging
import os
import re
from math import ceil

# import multiprocess as mp
from multiprocessing import pool
from pathlib import Path
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd
from spectrum_fundamentals.mod_string import xisearch_to_internal

from .search_results import SearchResults

logger = logging.getLogger(__name__)


class Xisearch(SearchResults):
    """Handle search results from xisearch."""

    def read_result(
        self,
        tmt_label: str = "",
        custom_mods: dict[str, int] | None = None,
        ptm_unimod_id: int | None = 0,
        ptm_sites: list[str] | None = None,
    ) -> pd.DataFrame:
        """
        Function to read a csv of CSMs and perform some basic formatting.

        :param tmt_label: tmt label as str
        :param custom_mods: dict with custom variable and static identifier and respecitve internal equivalent and mass
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        :raises NotImplementedError: if TMT label is provided
        :return: pd.DataFrame with the formatted data
        """
        if tmt_label != "":
            raise NotImplementedError("TMT is not supported for XIsearch")

        logger.info("Reading search results file...")
        columns_to_read = [
            "run_name",
            "scan_number",
            "precursor_mass",
            "precursor_charge",
            "crosslinker_name",
            "decoy_p1",
            "base_sequence_p1",
            "sequence_p1",
            "sequence_p2",
            "start_pos_p1",
            "start_pos_p2",
            "aa_len_p1",
            "link_pos_p1",
            "linked_aa_p1",
            "mods_p1",
            "mod_pos_p1",
            "protein_p1",
            "decoy_p2",
            "base_sequence_p2",
            "aa_len_p2",
            "link_pos_p2",
            "linked_aa_p2",
            "mods_p2",
            "mod_pos_p2",
            "protein_p2",
            "linear",
            "match_score",
        ]

        converters = {
            "mods_p1": str,
            "mods_p2": str,
            "mod_pos_p1": str,
            "mod_pos_p2": str,
            "start_pos_p1": str,
            "start_pos_p2": str,
        }

        self.results = pd.read_csv(self.path, sep="\t", usecols=columns_to_read, converters=converters)

        logger.info("Finished reading search results file.")
        # Standardize column names
        self.filter_xisearch_result()
        self.convert_to_internal(mods={})
        self.filter_valid_prosit_sequences()
        # df = Xisearch._filter_duplicates(df)
        self.results = Xisearch._fdr_group(self.results, fdr_group_col=None)

        return self.results

    def filter_xisearch_result(self):
        """Remove unsupported modifications and keep only k-k as linked amino acid."""
        df = self.results
        df["linear"] = df["linear"].fillna(True)
        df["linear"] = df["linear"].astype(bool)
        df = df[~df["linear"]]
        df = df[df["linked_aa_p1"].notna() & df["linked_aa_p1"].str.contains("K")]
        df = df[df["linked_aa_p2"].notna() & df["linked_aa_p2"].str.contains("K")]
        df = df[~df["mods_p1"].str.contains("dsso-hyd", na=False)]
        df = df[~df["mods_p2"].str.contains("dsso-hyd", na=False)]

        self.results = df

    @staticmethod
    def _self_or_between(df):
        df.loc[:, "protein_p1_arr"] = (
            df.loc[:, "protein_p1"].astype(str).str.replace("REV_", "").str.split(";").map(np.unique).map(list)
        )
        df.loc[:, "protein_p2_arr"] = (
            df.loc[:, "protein_p2"].astype(str).str.replace("REV_", "").str.split(";").map(np.unique).map(list)
        )
        df.loc[:, "proteins_arr"] = df.loc[:, "protein_p1_arr"] + df.loc[:, "protein_p2_arr"]
        df.loc[:, "proteins_arr_unique"] = df.loc[:, "proteins_arr"].map(np.unique)
        df.loc[:, "is_between"] = ~(
            df.loc[:, "proteins_arr"].map(len) - df.loc[:, "proteins_arr_unique"].map(len)
        ).astype(bool)
        df.drop(["protein_p1_arr", "protein_p2_arr", "proteins_arr", "proteins_arr_unique"], axis=1, inplace=True)
        return df.loc[:, "is_between"].map({True: "between", False: "self"})

    @staticmethod
    def _self_or_between_mp(df):
        pool_size = min([10, os.cpu_count()])
        slice_size = ceil(len(df) / pool_size)
        cols = ["protein_p1", "protein_p2"]
        df = df[cols]
        df_slices = [df.iloc[i * slice_size : (i + 1) * slice_size][cols] for i in range(pool_size)]
        print("slicing done")
        print(f"Pool size: {pool_size}")
        with pool.Pool(processes=pool_size) as self_or_between_pool:
            map_res = self_or_between_pool.map(Xisearch._self_or_between, df_slices)
        return pd.concat(map_res).copy()

    @staticmethod
    def _fdr_group(df, fdr_group_col=None):
        if fdr_group_col is None:
            df.loc[:, "fdr_group"] = Xisearch._self_or_between_mp(df)
        return df

    @staticmethod
    def _filter_duplicates(df: pd.DataFrame) -> pd.DataFrame:
        """
        Keep csm with higher score and remove duplicate (only top ranks).

        :param df: df to filter
        :return: filtered df as pd.DataFrame
        """
        repetitive_combinations = df[df.duplicated(subset=["scan_number", "run_name"], keep=False)]
        filtered_df = repetitive_combinations.groupby(["scan_number", "run_name"]).apply(
            lambda x: x.loc[x["match_score"].idxmax()]
        )
        filtered_df.reset_index(drop=True, inplace=True)
        final_df = pd.concat([df.drop_duplicates(subset=["scan_number", "run_name"], keep=False), filtered_df])
        final_df.reset_index(drop=True, inplace=True)
        df = final_df
        return df

    def convert_to_internal(
        self, mods: dict[str, str], ptm_unimod_id: int | None = None, ptm_sites: list[str] | None = None
    ):
        """
        Convert all columns in the search engine-specific output to the internal format used by Oktoberfest.

        :param mods: dictionary mapping search engine-specific mod patterns (keys) to ProForma standard (values)
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        """
        df = self.results
        df["crosslinker_name"] = df["crosslinker_name"].replace(to_replace="*", value="DSSO")
        df["decoy"] = df["decoy_p1"] | df["decoy_p2"]
        df["run_name"] = df["run_name"].str.replace("-", "_")
        df["RAW_FILE"] = df["run_name"]
        df["MASS"] = df["precursor_mass"]
        df["PRECURSOR_CHARGE"] = df["precursor_charge"]
        df["CROSSLINKER_TYPE"] = df["crosslinker_name"]
        df["SCORE"] = df["match_score"]
        df["REVERSE"] = df["decoy"]
        df["SCAN_NUMBER"] = df["scan_number"]
        df["SEQUENCE_A"] = df["base_sequence_p1"]
        df["SEQUENCE_B"] = df["base_sequence_p2"]
        df["Modifications_A"] = df["mods_p1"]
        df["Modifications_B"] = df["mods_p2"]
        df["CROSSLINKER_POSITION_A"] = df["link_pos_p1"]
        df["CROSSLINKER_POSITION_B"] = df["link_pos_p2"]
        df["ModificationPositions1"] = df["mod_pos_p1"]
        df["ModificationPositions2"] = df["mod_pos_p2"]
        df["PEPTIDE_LENGTH_A"] = df["aa_len_p1"]
        df["PEPTIDE_LENGTH_B"] = df["aa_len_p2"]
        logger.info("Converting Xisearch peptide sequence to internal format...")

        df["RAW_FILE"] = df["RAW_FILE"].str.replace(".raw", "", regex=False)

        df["MODIFIED_SEQUENCE_A"] = df.apply(
            lambda row: xisearch_to_internal(
                xl=row["CROSSLINKER_TYPE"],
                seq=row["SEQUENCE_A"],
                mod=row["Modifications_A"],
                crosslinker_position=row["CROSSLINKER_POSITION_A"],
                mod_positions=row["ModificationPositions1"],
            ),
            axis=1,
            result_type="expand",
        )

        df["MODIFIED_SEQUENCE_B"] = df.apply(
            lambda row: xisearch_to_internal(
                xl=row["CROSSLINKER_TYPE"],
                seq=row["SEQUENCE_B"],
                mod=row["Modifications_B"],
                crosslinker_position=row["CROSSLINKER_POSITION_B"],
                mod_positions=row["ModificationPositions2"],
            ),
            axis=1,
            result_type="expand",
        )

        self.results = df

    def filter_valid_prosit_sequences(self) -> pd.DataFrame:
        """
        Filter valid Prosit sequences.

        :return: df after filtration
        """
        logger.info(f"#sequences before filtering for valid prosit sequences: {len(self.results)}")
        df = self.results
        df = df[(df["PEPTIDE_LENGTH_A"] <= 30)]
        df = df[df["PEPTIDE_LENGTH_A"] >= 6]
        df = df[(df["PEPTIDE_LENGTH_B"] <= 30)]
        df = df[df["PEPTIDE_LENGTH_B"] >= 6]
        df = df[(~df["SEQUENCE_A"].str.contains(r"B|\*|\.|U|O|X|Z|\(|\)"))]
        df = df[(~df["SEQUENCE_B"].str.contains(r"B|\*|\.|U|O|X|Z|\(|\)"))]
        df = df[df["PRECURSOR_CHARGE"] <= 6]
        logger.info(f"#sequences after filtering for valid prosit sequences: {len(df)}")
        self.results = df
        return self.results



"""SpectrumIO: File / Data Conversion for Mass Spec data within the Oktoberfest ecosystem."""

from datetime import datetime

__author__ = """The Oktoberfest development team (Wilhelmlab at Technical University of Munich)"""
__copyright__ = f"Copyright {datetime.now():%Y}, Wilhelmlab at Technical University of Munich"
__license__ = "MIT"
__version__ = "0.6.4"

import logging
import logging.handlers
import sys

from spectrum_io import d, file, raw

# from .search_result import MaxQuant
# from .spectral_library import DLib, Spectronaut

CONSOLE_LOG_LEVEL = logging.INFO
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class _InfoWarningFilter(logging.Filter):
    def filter(self, record):
        return CONSOLE_LOG_LEVEL <= record.levelno <= logging.WARNING


if len(logger.handlers) == 0:
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s::%(funcName)s %(message)s")
    # add console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.addFilter(_InfoWarningFilter())
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    # add error handler
    error_handler = logging.StreamHandler()
    error_handler.setLevel(logging.ERROR)
    error_handler.setFormatter(formatter)
    logger.addHandler(error_handler)
else:
    logger.info("Logger already initialized. Resuming normal operation.")



import re
from abc import abstractmethod
from multiprocessing import Queue
from multiprocessing.managers import ValueProxy
from pathlib import Path
from sqlite3 import Connection
from typing import IO, Dict, Optional, Union

import numpy as np
import pandas as pd


def parse_mods(mods: Dict[str, int]) -> Dict[str, str]:
    """
    Parse provided mapping of custom modification pattern to ProForma standard.

    This function takes a dictionary mapping custom modification pattern for specific aminoacids (keys) to a
    UNIMOD ID (values). The pattern is translated to ProForma standard and a new dictionary mapping the custom
    modification patterns to the ProForma standard is returned.
    The pattern for the custom modifications must start with the one-letter code for an aminoacid or '^' / '$',
    to describe n- / c-terminal modifications, respectively, followed by an optional pattern (which can be
    empty).
    This means that 'X'  or 'X(custom_pattern)', is both mapped to 'X[UNIMOD:#]'.
    For the n-terminus, an additional dash will be added automatically, which maps 'X(custom_pattern)' to
    'X[UNIMOD:#]-'. If the sequence to apply the transformation on already contains the dash, it needs to be part
    of the custom_pattern (i.e. 'X(custom_pattern)-'), to avoid adding an additional dash.

    :param mods: Dictionary mapping custom modification patterns (keys) to UNIMOD IDs (values)
    :raises TypeError: if keys are not strings or values are not integers
    :raises ValueError: if the keys do not start with [A-Z,a-z,^,$]
    :return: A dictionary mapping custom modification patterns (keys) to the ProForma standard (values)
    """
    key_pattern = (
        "'X' or 'X<mod_pattern>' where X is either the one-letter code of an aminoacid or '^' / '$' defining the"
        " n- or c-terminus, respectively, followed by an optional pattern identifying a specific modification."
    )
    unimod_regex_map = {}
    for k, v in mods.items():
        if not isinstance(v, int):
            raise TypeError(f"UNIMOD id {v} for replacement {k} not understood. UNIMOD IDs must be integers.")
        if not isinstance(k, str):
            raise TypeError(
                f"Replacement {k} not understood. Replacements must be strings and follow " f"the pattern {key_pattern}"
            )
        if k[0].isalpha():
            unimod_regex_map[re.escape(f"{k[0]}[UNIMOD:{v}]")] = k
            continue
        if k[0] == "^":
            to_escape = re.escape(f"[UNIMOD:{v}]-")
            unimod_regex_map[f"^{to_escape}"] = k[1:]
            continue
        raise ValueError(
            f"Replacement {k} not understood. {k[0]} is not a valid aminoacid. "
            f"Replacements most follow the pattern {key_pattern}"
        )
    return unimod_regex_map


class SpectralLibrary:
    """Main to initialze a SpectralLibrary obj."""

    @property
    @abstractmethod
    def standard_mods(self) -> Dict[str, int]:
        """Standard modifications that are always applied if not otherwise specified."""
        pass

    def __init__(
        self,
        output_path: Union[str, Path],
        mode: str = "w",
        min_intensity_threshold: float = 5e-4,
        chunksize: Optional[int] = None,
    ):
        """
        Initialize a SpectralLibrary obj.

        :param output_path: path to output file including file name
        :param mode: Whether to append ('a') to or overwrite ('w') an extisting file
            at the provided output path (if it is present).
        :param min_intensity_threshold: optional filter for low intensity peaks
        :param chunksize: optional chunksize for dlib
        """
        if isinstance(output_path, str):
            output_path = Path(output_path)
        self.out_path = output_path
        self.mode = mode
        self.min_intensity_threshold = min_intensity_threshold
        self.chunksize = chunksize

    def load(self):
        """Load predictions from hdf5 file."""

    def write(self, *, custom_mods: Optional[Dict[str, int]] = None, **kwargs):
        """
        Write content to the output file.

        :param custom_mods: optional dictionary mapping libary format-specific modification patterns (keys)
            to UNIMOD IDs (values)
        :param kwargs: Keyword arguments to be passed to the internal _write method.
        """
        parsed_mods = parse_mods(self.standard_mods | (custom_mods or {}))
        with self._get_handle() as out:
            self._initialize(out)
            self._write(out, mods=parsed_mods, **kwargs)

    def _get_handle(self):
        return open(self.out_path, self.mode)

    def async_write(self, queue: Queue, progress: ValueProxy, custom_mods: Optional[Dict[str, int]] = None):
        """
        Asynchronously write content to the output file from a queue.

        :param queue: A queue from which content will be retrieved for writing.
        :param progress: An integer value representing the progress of the writing process.
        :param custom_mods: dict with custom variable and static identifier and respecitve internal equivalent and mass
        """
        parsed_mods = parse_mods(self.standard_mods | (custom_mods or {}))

        with self._get_handle() as out:
            self._initialize(out)
            while True:
                content = queue.get()
                if content is None:
                    break
                data, metadata = content
                self._write(out, data=data, metadata=metadata, mods=parsed_mods)
                progress.value += 1

    def _fragment_filter_passed(
        self, f_mz: Union[np.ndarray, float], f_int: Union[np.ndarray, float]
    ) -> Union[np.ndarray, bool]:
        """
        Return if a fragment passes a common filter.

        This function returns a boolean to determine if a fragment can exist (as defined by its mz != -1)
        and if its (predicted) intensity is at least the minimal threshold given during instantiation of
        the writer class.

        :param f_mz: The fragment's mass to charge ratio
        :param f_int: The fragment's (predicted) intensity

        :return: boolean (array) that determines if the fragment passes the filter
        """
        return (f_mz != -1) & (f_int >= self.min_intensity_threshold)

    @abstractmethod
    def _write(
        self,
        out: Union[IO, Connection],
        data: Dict[str, np.ndarray],
        metadata: pd.DataFrame,
        mods: Dict[str, str],
    ):
        """
        Internal writer function.

        This function takes a batch of data and corresponding metadata and writes it to the provided
        output file handle. The file handle logic is provided using the public write or async_write
        functions.

        :param out: file handle accepting the data to be written to disk
        :param data: Dictionary containing TODO keys and corresponding values as numpy array
        :param metadata: a dataframe that contains the columns TODO
        :param mods: optional dictionary mapping libary format-specific modification patterns (keys)
            to UNIMOD IDs (values)

        """
        pass

    @abstractmethod
    def _initialize(self, out: Union[IO, Connection]):
        pass



from __future__ import annotations

import logging
from typing import Dict, Optional

import pandas as pd
import spectrum_fundamentals.constants as c
from pyteomics import pepxml
from spectrum_fundamentals.constants import MSFRAGGER_VAR_MODS
from spectrum_fundamentals.mod_string import add_permutations, internal_without_mods
from tqdm import tqdm

from .search_results import SearchResults, parse_mods

logger = logging.getLogger(__name__)


class MSFragger(SearchResults):
    """Handle search results from MSFragger."""

    @property
    def standard_mods(self):
        """Standard modifications that are always applied if not otherwise specified."""
        return {"C[160]": 4, "M[147]": 35, "R[157]": 7, "Q[129]": 7, "N[115]": 7}

    def filter_valid_prosit_sequences(self):
        """Filter valid Prosit sequences."""
        logger.info(f"#sequences before filtering for valid prosit sequences: {len(self.results.index)}")
        # retain only peptides that fall within [7, 30] length supported by Prosit
        self.results = self.results[(self.results["PEPTIDE_LENGTH"] <= 30) & (self.results["PEPTIDE_LENGTH"] >= 7)]
        # remove unsupported mods to exclude
        self.results = self.results[~self.results["MODIFIED_SEQUENCE"].str.contains(r"\[\d+\]", regex=True)]
        # remove precursor charges greater than 6
        self.results = self.results[self.results["PRECURSOR_CHARGE"] <= 6]
        logger.info(f"#sequences after filtering for valid prosit sequences: {len(self.results.index)}")

        return self.results

    def read_result(
        self,
        tmt_label: str = "",
        custom_mods: dict[str, int] | None = None,
        ptm_unimod_id: int | None = 0,
        ptm_sites: list[str] | None = None,
    ) -> pd.DataFrame:
        """
        Function to read a msms txt and perform some basic formatting.

        :param tmt_label: optional tmt label as str
        :param custom_mods: optional dictionary mapping MSFragger-specific mod pattern to UNIMOD IDs.
            If None, static carbamidomethylation of cytein and variable oxidation of methionine
            are mapped automatically. To avoid this, explicitely provide an empty dictionary.
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        :raises FileNotFoundError: in case the given path is neither a file, nor a directory.
        :return: pd.DataFrame with the formatted data
        """
        parsed_mods = parse_mods(self.standard_mods | (custom_mods or {}))
        if tmt_label:
            unimod_tag = c.TMT_MODS[tmt_label]
            parsed_mods["K"] = f"K{unimod_tag}"
            parsed_mods[r"^n\[\d+\]"] = f"{unimod_tag}-"
        if self.path.is_file():
            file_list = [self.path]
        elif self.path.is_dir():
            file_list = list(self.path.rglob("*.pepXML"))
        else:
            raise FileNotFoundError(f"{self.path} could not be found.")

        ms_frag_results = []
        for pep_xml_file in tqdm(file_list):
            ms_frag_results.append(pepxml.DataFrame(str(pep_xml_file)))

        self.results = pd.concat(ms_frag_results)

        self.convert_to_internal(mods=parsed_mods, ptm_unimod_id=ptm_unimod_id, ptm_sites=ptm_sites)
        return self.filter_valid_prosit_sequences()

    @staticmethod
    def check_decoys(protein_names: str):
        """
        Check if all protein names in a given string correspond to decoy proteins.

        :param protein_names: A string containing one or more protein names separated by semicolons (';').
                            Each protein name is checked for the presence of the substring 'rev'.
        :return: `True` if all proteins are decoy proteins (i.e., if all protein names contain 'rev'),
                otherwise `False`.
        """
        all_proteins = protein_names.split(";")
        reverse = True
        for protein in all_proteins:
            if "rev" not in protein:
                reverse = False
                break
        return reverse

    def convert_to_internal(self, mods: dict[str, str], ptm_unimod_id: int | None, ptm_sites: list[str] | None):
        """
        Convert all columns in the MSFragger output to the internal format used by Oktoberfest.

        :param mods: dictionary mapping MSFragger-specific mod patterns (keys) to ProForma standard (values)
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        """
        df = self.results
        df["protein"] = df["protein"].fillna("UNKNOWN").apply(lambda x: ";".join(x))

        df["REVERSE"] = df["protein"].apply(lambda x: MSFragger.check_decoys(x))

        df["spectrum"] = df["spectrum"].str.split(pat=".", n=1).str[0]
        df["PEPTIDE_LENGTH"] = df["peptide"].str.len()
        df.replace({"modified_peptide": mods}, regex=True, inplace=True)
        df["peptide"] = internal_without_mods(df["modified_peptide"])
        if ptm_unimod_id != 0:

            # PTM permutation generation
            if ptm_unimod_id == 7:
                allow_one_less_modification = True
            else:
                allow_one_less_modification = False

            df["modified_peptide"] = df["modified_peptide"].apply(
                add_permutations,
                unimod_id=ptm_unimod_id,
                residues=ptm_sites,
                allow_one_less_modification=allow_one_less_modification,
            )
            df = df.explode("modified_peptide", ignore_index=True)

        df.rename(
            columns={
                "assumed_charge": "PRECURSOR_CHARGE",
                "index": "SCAN_EVENT_NUMBER",
                "peptide": "SEQUENCE",
                "start_scan": "SCAN_NUMBER",
                "hyperscore": "SCORE",
                "modified_peptide": "MODIFIED_SEQUENCE",
                "protein": "PROTEINS",
                "peptide": "SEQUENCE",
                "precursor_neutral_mass": "MASS",
                "spectrum": "RAW_FILE",
            },
            inplace=True,
        )
        self.results = df
        """
        return df[
            [
                "RAW_FILE",
                "SCAN_NUMBER",
                "MODIFIED_SEQUENCE",
                "PRECURSOR_CHARGE",
                "SCAN_EVENT_NUMBER",
                "MASS",
                "SCORE",
                "REVERSE",
                "SEQUENCE",
                "PEPTIDE_LENGTH",
                "PROTEINS",
            ]
        ]
        """



from __future__ import annotations

import glob
import logging
import os
import re
from pathlib import Path
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd
import spectrum_fundamentals.constants as c
from spectrum_fundamentals.mod_string import xisearch_to_internal

from .search_results import SearchResults

logger = logging.getLogger(__name__)


class Scout(SearchResults):
    """Handle search results from xisearch."""

    def read_result(
        self,
        tmt_label: str = "",
        custom_mods: dict[str, int] | None = None,
        ptm_unimod_id: int | None = 0,
        ptm_sites: list[str] | None = None,
    ) -> pd.DataFrame:
        """
        Function to read a csv of CSMs and perform some basic formatting.

        :param tmt_label: tmt label as str
        :param custom_mods: dict with custom variable and static identifier and respecitve internal equivalent and mass
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        :raises NotImplementedError: if TMT label is provided
        :return: pd.DataFrame with the formatted data
        """
        if tmt_label != "":
            raise NotImplementedError("TMT is not supported for Scout")

        logger.info("Reading search results file...")
        columns_to_read = [
            "ScanNumber",
            "Charge",
            "ExperimentalMZ",
            "AlphaPeptide",
            "BetaPeptide",
            "AlphaPos",
            "BetaPos",
            "AlphaMappings",
            "BetaMappings",
            "ClassificationScore",
            "Peptide Position 1",
            "Peptide Position 2",
            "Protein 1",
            "Protein 2",
            "FileName",
        ]

        self.results = pd.read_csv(self.path, usecols=columns_to_read)
        logger.info("Finished reading search results file.")
        # Standardize column names
        self.convert_to_internal(mods={})
        return self.filter_valid_prosit_sequences()
        # df = Scout._filter_duplicates(df)

    @staticmethod
    def _filter_duplicates(df: pd.DataFrame) -> pd.DataFrame:
        """
        Keep csm with higher score and remove duplicate (only top ranks).

        :param df: df to filter
        :return: filtered df as pd.DataFrame
        """
        repetitive_combinations = df[df.duplicated(subset=["ScanNumber", "RAW_FILE"], keep=False)]
        filtered_df = repetitive_combinations.groupby(["ScanNumber", "RAW_FILE"]).apply(
            lambda x: x.loc[x["ClassificationScore"].idxmax()]
        )
        filtered_df.reset_index(drop=True, inplace=True)
        final_df = pd.concat([df.drop_duplicates(subset=["ScanNumber", "RAW_FILE"], keep=False), filtered_df])
        final_df.reset_index(drop=True, inplace=True)
        df = final_df
        return df

    @staticmethod
    def _extract_modifications(peptide_seq: str):
        modifications = []
        # Find all matches of modifications
        matches = re.findall(r"([CM])\(\+([\d.]+)\)", peptide_seq)
        for match in matches:
            mod, _ = match
            # Add modification to the list
            if mod == "C":
                modifications.append("cm")
            elif mod == "M":
                modifications.append("ox")
        return ";".join(modifications)

    @staticmethod
    def _extract_modification_positions(peptide_seq: str):
        pattern = r"([A-Z])(\(\+\d+\.\d+\))?"
        matches = re.findall(pattern, peptide_seq)
        split_peptide = []
        for match in matches:
            amino_acid = match[0]
            modification = match[1] if match[1] else ""
            split_peptide.append(amino_acid + modification)
        positions = [str(i + 1) for i, component in enumerate(split_peptide) if "+" in component]
        return ";".join(positions)

    @staticmethod
    def _self_or_between_mp(df: pd.DataFrame) -> pd.DataFrame:
        df["tmp_id"] = df.index
        df_expl = df.copy()
        df_expl.loc[:, "AlphaMappings"] = df_expl["AlphaMappings"].str.split(";")
        df_expl.loc[:, "BetaMappings"] = df_expl["BetaMappings"].str.split(";")
        df_expl = df_expl.explode("AlphaMappings")
        df_expl = df_expl.explode("BetaMappings")
        df_expl.loc[:, "self"] = False
        df_expl.loc[df_expl["AlphaMappings"] == df_expl["BetaMappings"], "self"] = True
        id_to_self = df_expl.groupby("tmp_id", dropna=False).agg({"self": "max"}).reset_index()
        df = df.drop(["self"], axis=1, errors="ignore").merge(id_to_self, on=["tmp_id"], validate="1:1")
        df.loc[:, "fdr_group"] = df["self"].apply(lambda x: "self" if x else "between")
        return df

    def convert_to_internal(
        self, mods: dict[str, str], ptm_unimod_id: int | None = None, ptm_sites: list[str] | None = None
    ):
        """
        Convert all columns in the search engine-specific output to the internal format used by Oktoberfest.

        :param mods: dictionary mapping search engine-specific mod patterns (keys) to ProForma standard (values)
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        """
        # Filter csms that does not contain any "k"
        df = self.results
        df = df[(df["AlphaPeptide"].str.contains("K")) & (df["BetaPeptide"].str.contains("K"))]
        df["decoy_p1"] = df["AlphaMappings"].str.contains("Reverse").astype(bool)
        df["decoy_p2"] = df["BetaMappings"].str.contains("Reverse").astype(bool)
        df["protein_p1"] = df["AlphaMappings"]
        df["protein_p2"] = df["BetaMappings"]
        df["decoy"] = df["decoy_p1"] | df["decoy_p2"]
        df["REVERSE"] = df["decoy"]
        df["RAW_FILE"] = df["FileName"].apply(lambda x: x.split("\\")[-1])
        df["MASS"] = df["ExperimentalMZ"]
        df["PRECURSOR_CHARGE"] = df["Charge"]
        df["CROSSLINKER_TYPE"] = "DSSO"
        df["crosslinker_name"] = "DSSO"
        df["linked_aa_p1"] = "K"
        df["linked_aa_p2"] = "K"
        df["linear"] = "False"
        df["match_score"] = "ClassificationScore"
        df["SCORE"] = df["ClassificationScore"]
        df["SCAN_NUMBER"] = df["ScanNumber"]
        df["SEQUENCE_A"] = df["AlphaPeptide"].apply(lambda x: re.sub(r"\([^)]*\)", "", x))
        df["SEQUENCE_B"] = df["BetaPeptide"].apply(lambda x: re.sub(r"\([^)]*\)", "", x))
        df["base_sequence_p1"] = df["SEQUENCE_A"]
        df["base_sequence_p2"] = df["SEQUENCE_B"]
        df = df[df.apply(lambda row: row["SEQUENCE_A"][row["AlphaPos"]] == "K", axis=1)]
        df = df[df.apply(lambda row: row["SEQUENCE_B"][row["BetaPos"]] == "K", axis=1)]
        df["Modifications_A"] = df["AlphaPeptide"].apply(Scout._extract_modifications)
        df["Modifications_B"] = df["BetaPeptide"].apply(Scout._extract_modifications)
        df["mods_p1"] = df["AlphaPeptide"].apply(Scout._extract_modifications)
        df["mods_p2"] = df["BetaPeptide"].apply(Scout._extract_modifications)
        df["ModificationPositions1"] = df["AlphaPeptide"].apply(Scout._extract_modification_positions)
        df["ModificationPositions2"] = df["BetaPeptide"].apply(Scout._extract_modification_positions)
        df["CROSSLINKER_POSITION_A"] = df["AlphaPos"] + 1
        df["CROSSLINKER_POSITION_B"] = df["BetaPos"] + 1
        df["mod_pos_p1"] = df["AlphaPos"] + 1
        df["mod_pos_p2"] = df["BetaPos"] + 1
        df["link_pos_p1"] = df["AlphaPos"] + 1
        df["link_pos_p2"] = df["BetaPos"] + 1
        df["PEPTIDE_LENGTH_A"] = df["SEQUENCE_A"].apply(len)
        df["PEPTIDE_LENGTH_B"] = df["SEQUENCE_B"].apply(len)
        df["aa_len_p1"] = df["SEQUENCE_A"].apply(len)
        df["aa_len_p2"] = df["SEQUENCE_B"].apply(len)
        df = Scout._self_or_between_mp(df)
        df["fdr_group"] = np.where(
            df["AlphaMappings"].str.replace("Reverse_", "") == df["BetaMappings"].str.replace("Reverse_", ""),
            "self",
            "between",
        )
        df.drop(columns=["self"], inplace=True)
        df.drop(columns=["tmp_id"], inplace=True)
        logger.info("Converting Scout peptide sequence to internal format...")
        df["RAW_FILE"] = df["RAW_FILE"].str.replace(".raw", "")
        df["MODIFIED_SEQUENCE_A"] = df.apply(
            lambda row: xisearch_to_internal(
                xl=row["CROSSLINKER_TYPE"],
                seq=row["SEQUENCE_A"],
                mod=row["Modifications_A"],
                crosslinker_position=row["CROSSLINKER_POSITION_A"],
                mod_positions=row["ModificationPositions1"],
            ),
            axis=1,
            result_type="expand",
        )
        df["MODIFIED_SEQUENCE_B"] = df.apply(
            lambda row: xisearch_to_internal(
                xl=row["CROSSLINKER_TYPE"],
                seq=row["SEQUENCE_B"],
                mod=row["Modifications_B"],
                crosslinker_position=row["CROSSLINKER_POSITION_B"],
                mod_positions=row["ModificationPositions2"],
            ),
            axis=1,
            result_type="expand",
        )
        new_column_names = {
            "FileName": "run_name",
            "ScanNumber": "scan_number",
            "ExperimentalMZ": "precursor_mass",
            "Charge": "precursor_charge",
            "scan_number": "ScanNumber",
        }
        self.results = df.rename(columns=new_column_names)

    def filter_valid_prosit_sequences(self) -> pd.DataFrame:
        """
        Filter valid Prosit sequences.

        :return: df after filtration
        """
        logger.info(f"#sequences before filtering for valid prosit sequences: {len(self.results)}")
        self.results = self.results[(self.results["PEPTIDE_LENGTH_A"] <= 30)]
        self.results = self.results[self.results["PEPTIDE_LENGTH_A"] >= 6]
        self.results = self.results[(self.results["PEPTIDE_LENGTH_B"] <= 30)]
        self.results = self.results[self.results["PEPTIDE_LENGTH_B"] >= 6]
        self.results = self.results[(~self.results["SEQUENCE_A"].str.contains(r"B|\*|\.|U|O|X|Z|\(|\)"))]
        self.results = self.results[(~self.results["SEQUENCE_B"].str.contains(r"B|\*|\.|U|O|X|Z|\(|\)"))]
        self.results = self.results[self.results["PRECURSOR_CHARGE"] <= 6]
        logger.info(f"#sequences after filtering for valid prosit sequences: {len(self.results)}")

        return self.results



from __future__ import annotations

import logging
import re
from abc import abstractmethod
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

import pandas as pd

from spectrum_io.file import csv

logger = logging.getLogger(__name__)


COLUMNS = [
    "RAW_FILE",
    "SCAN_NUMBER",
    "MODIFIED_SEQUENCE",
    "PRECURSOR_CHARGE",
    "MASS",
    "SCORE",
    "REVERSE",
    "SEQUENCE",
    "PEPTIDE_LENGTH",
    "PROTEINS",
]


def parse_mods(mods: dict[str, int]) -> dict[str, str]:
    """
    Parse provided mapping of custom modification pattern to ProForma standard.

    This function takes a dictionary mapping custom modification pattern for specific aminoacids (keys) to a
    UNIMOD ID (values). The pattern is translated to ProForma standard and a new dictionary mapping the custom
    modification patterns to the ProForma standard is returned.
    The pattern for the custom modifications must start with the one-letter code for an aminoacid or '^' / '$',
    to describe n- / c-terminal modifications, respectively, followed by an optional pattern (which can be
    empty).
    This means that 'X'  or 'X(custom_pattern)', is both mapped to 'X[UNIMOD:#]'.
    For the n-terminus, an additional dash will be added automatically, which maps 'X(custom_pattern)' to
    'X[UNIMOD:#]-'. If the sequence to apply the transformation on already contains the dash, it needs to be part
    of the custom_pattern (i.e. 'X(custom_pattern)-'), to avoid adding an additional dash.

    :param mods: Dictionary mapping custom modification patterns (keys) to UNIMOD IDs (values)
    :raises TypeError: if keys are not strings or values are not integers
    :raises ValueError: if the keys do not start with [A-Z,a-z,^,$]
    :return: A dictionary mapping custom modification patterns (keys) to the ProForma standard (values)
    """
    key_pattern = (
        "'X' or 'X<mod_pattern>' where X is either the one-letter code of an aminoacid or '^' / '$' defining the"
        " n- or c-terminus, respectively, followed by an optional pattern identifying a specific modification."
    )
    unimod_regex_map = {}
    for k, v in mods.items():
        if not isinstance(v, int):
            raise TypeError(f"UNIMOD id {v} for replacement {k} not understood. UNIMOD IDs must be integers.")
        if not isinstance(k, str):
            raise TypeError(
                f"Replacement {k} not understood. Replacements must be strings and follow " f"the pattern {key_pattern}"
            )
        if k[0].isalpha():
            unimod_regex_map[re.escape(k)] = f"{k[0].upper()}[UNIMOD:{v}]"
            continue
        if k[0] == "^":
            unimod_regex_map[f"^{re.escape(k[1:])}"] = f"[UNIMOD:{v}]-"
            continue
        raise ValueError(
            f"Replacement {k} not understood. {k[0]} is not a valid aminoacid. "
            f"Replacements most follow the pattern {key_pattern}"
        )
    # sort descending by length to avoid replacement of single-aa keys within larger keys
    return dict(sorted(unimod_regex_map.items(), key=lambda x: -len(x[0])))


class SearchResults:
    """Handle search results from different software."""

    orig_res: pd.DataFrame
    fake_msms: pd.DataFrame

    def __init__(self, path: str | Path):
        """
        Init Searchresults object.

        :param path: path to file
        """
        if isinstance(path, str):
            path = Path(path)
        self.path = path

    @abstractmethod
    def filter_valid_prosit_sequences(self):
        """Filter valid Prosit sequences."""
        raise NotImplementedError

    @abstractmethod
    def read_result(
        self,
        tmt_label: str = "",
        custom_mods: dict[str, int] | None = None,
        ptm_unimod_id: int | None = 0,
        ptm_sites: list[str] | None = None,
    ):
        """Read result.

        :param tmt_label: tmt label as str
        :param custom_mods: dict with custom variable and static identifier and respecitve internal equivalent and mass
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        """
        raise NotImplementedError

    def generate_internal(
        self,
        tmt_label: str = "",
        out_path: str | Path | None = None,
        custom_mods: dict[str, int] | None = None,
        ptm_unimod_id: int | None = 0,
        ptm_sites: list[str] | None = None,
        xl: bool = False,
    ) -> pd.DataFrame:
        """
        Generate df and save to out_path if provided.

        :param out_path: path to output
        :param tmt_label: tmt label as str
        :param custom_mods: dict with static and variable custom modifications, their internal identifier and mass
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        :param xl: set to True for crosslinking data
        :return: path to output file
        """
        if out_path is None:
            # convert and return
            filtered_df = self.read_result(
                tmt_label, custom_mods=custom_mods, ptm_unimod_id=ptm_unimod_id, ptm_sites=ptm_sites
            )
            if xl:
                return filtered_df[:]  # Return all columns
            else:
                return filtered_df[COLUMNS]
        if isinstance(out_path, str):
            out_path = Path(out_path)

        if out_path.is_file():
            # only read converted and return
            logger.info(f"Found search results in internal format at {out_path}, skipping conversion")
            # TODO: internal_to_unimod
            return csv.read_file(out_path)

        # convert, save and return
        if xl:
            df = self.read_result(tmt_label, custom_mods=custom_mods, ptm_unimod_id=ptm_unimod_id, ptm_sites=ptm_sites)[
                :
            ]
        else:
            df = self.read_result(tmt_label, custom_mods=custom_mods, ptm_unimod_id=ptm_unimod_id, ptm_sites=ptm_sites)[
                COLUMNS
            ]
        csv.write_file(df, out_path)
        return df

    def read_internal(self) -> pd.DataFrame:
        """
        Read file from path.

        :return: dataframe after reading the file
        """
        return csv.read_file(self.path)

    @abstractmethod
    def convert_to_internal(self, mods: dict[str, str], ptm_unimod_id: int | None, ptm_sites: list[str] | None):
        """
        Convert all columns in the search engine-specific output to the internal format used by Oktoberfest.

        :param mods: dictionary mapping search engine-specific mod patterns (keys) to ProForma standard (values)
        :param ptm_unimod_id: unimod id used for site localization
        :param ptm_sites: possible sites that the ptm can exist on
        """
        raise NotImplementedError



"""Initialize seach result."""

from .mascot import Mascot
from .maxquant import MaxQuant
from .msamanda import MSAmanda
from .msfragger import MSFragger
from .sage import Sage
from .scout import Scout
from .xisearch import Xisearch
